import numpy.typing as npt
import numpy as np
from kluster_fudge.dist import DistanceMetrics, distance
from kluster_fudge.init import init_centroids, InitMethod
from kluster_fudge.utils import update_centroids


class KModes:
    def __init__(
        self,
        n_clusters: int = 8,
        n_init: int = 10,
        max_iter: int = 100,
        init_method: str = "cao",
        dist_metric: str = "hamming",
        random_state: int = 42,
    ) -> None:
        self.n_clusters = n_clusters
        self.n_init = n_init
        self.max_iter = max_iter
        self.dist_metric = dist_metric
        self.centroids = None
        self.labels = None
        self.encodings = []
        self.decoded_centroids = None
        self.is_df = False
        self.random_state = random_state

        if init_method == "random":
            self.init_method = InitMethod.RAND
        elif init_method == "huang":
            self.init_method = InitMethod.HUANG
        elif init_method == "cao":
            self.init_method = InitMethod.CAO
        else:
            raise ValueError(f"Unknown init method: {init_method}")

        if dist_metric == "hamming":
            self.dist_metric = DistanceMetrics.HAMMING
        elif dist_metric == "jaccard":
            self.dist_metric = DistanceMetrics.JACCARD
        elif dist_metric == "ng":
            self.dist_metric = DistanceMetrics.NG
        else:
            raise ValueError(f"Unknown distance metric: {dist_metric}")

    def _encode(self, X: npt.NDArray) -> npt.NDArray[np.int64]:
        self.encodings = []
        X_encoded = np.zeros(X.shape, dtype=int)

        # every column has its own integer encoding
        # e.g. for mapping ["a", "b", "c"] -> [0, 1, 2], we have ["a", "b", "c", "a", "b"] -> [0, 1, 2, 0, 1]
        for i in range(X.shape[1]):
            unique_vals, encoded = np.unique(X[:, i], return_inverse=True)
            self.encodings.append(unique_vals)
            X_encoded[:, i] = encoded

        return X_encoded

    def _decode(self, centroids_encoded: npt.NDArray[np.int64]) -> npt.NDArray[np.str_]:
        """
        Decode the centroids from integer encoding to original values.

        Args:
            centroids_encoded: (npt.NDArray[np.int64]) Array of encoded centroids (n_clusters, n_features)

        Returns:
            (npt.NDArray[np.str_]) Decoded centroids array (n_clusters, n_features)
        """

        decoded = []
        for i in range(len(self.encodings)):
            col_map = self.encodings[i]
            decoded_col = col_map[centroids_encoded[:, i].astype(int)]
            decoded.append(decoded_col)

        # decoded is a list of arrays representing columns; stack horizontally so each row is a centroid
        return np.array(decoded).T

    def fit(self, X: npt.ArrayLike) -> None:
        """
        Fit the model to the input data.

        Args:
            X: (npt.ArrayLike) Data array (n_samples, n_features)

        Returns:
            None
        """

        # check if X is a pandas dataframe, if so, convert to numpy array
        if hasattr(X, "values"):
            X = X.values
            self.is_df = True
        else:
            X = np.asarray(X)

        # encode X into integer array for efficiency
        X = self._encode(X)
        self.centroids = init_centroids(
            X, self.n_clusters, self.init_method, random_state=self.random_state
        )

        self.labels = np.zeros(X.shape[0], dtype=int)
        for i in range(self.max_iter):  # for the number of iterations, fit, then adjust
            # Use Hamming for the first iteration if metric is NG to generate initial labels
            current_metric = self.dist_metric
            if i == 0 and self.dist_metric == DistanceMetrics.NG:
                current_metric = DistanceMetrics.HAMMING

            dist = distance(
                X, self.centroids, current_metric, labels=self.labels
            )  # compute distance

            # Assign each point to its closest centroid (vectorize using np.argmin on axis 1)
            self.labels = np.argmin(dist, axis=1)

            # update centroids
            self.centroids = update_centroids(X, self.labels, self.n_clusters)

        self.decoded_centroids = self._decode(self.centroids)

    def predict(self, X: npt.ArrayLike) -> npt.NDArray[np.int64]:
        """
        Predict the cluster labels for the input data.

        Args:
            X: (npt.ArrayLike) Data array (n_samples, n_features)

        Returns:
            (npt.NDArray[np.int64]) Labels array (n_samples,)
        """
        if hasattr(X, "values"):
            X = X.values
        else:
            X = np.asarray(X)

        X = self._encode(X)
        dist = distance(X, self.centroids, self.dist_metric)
        return np.argmin(dist, axis=1)

    def fit_predict(self, X: npt.ArrayLike) -> npt.NDArray[np.int64]:
        """
        Fit the model to the input data and return the cluster labels.

        Args:
            X: (npt.ArrayLike) Data array (n_samples, n_features)

        Returns:
            (npt.NDArray[np.int64]) Labels array (n_samples,)
        """
        self.fit(X)
        return self.labels



import numpy as np
import pytest
from kluster_fudge.dist import distance, hamming, jaccard, ng, DistanceMetrics


def test_hamming_simple():
    # dataset setup
    X = np.array([[0, 0], [0, 1], [1, 1]], dtype=np.int64)
    centroids = np.array([[0, 0], [1, 1]], dtype=np.int64)

    # expected diffs:
    # x0 vs c0: 0
    # x0 vs c1: 2
    # x1 vs c0: 1
    # ...

    dist = hamming(X, centroids)

    expected = np.array([[0, 2], [1, 1], [2, 0]], dtype=np.float64)

    np.testing.assert_array_equal(dist, expected)


def test_jaccard_simple():
    # dataset setup
    X = np.array([[0, 0]], dtype=np.int64)
    centroids = np.array([[0, 0], [1, 1]], dtype=np.int64)

    # x0 vs c0: match=2, total=2. dist=0
    # x0 vs c1: match=0, total=4. dist=1

    dist = jaccard(X, centroids)
    expected = np.array([[0.0, 1.0]])
    np.testing.assert_array_almost_equal(dist, expected)


def test_ng_simple():
    X = np.array([[0, 0], [0, 1]], dtype=np.int64)
    centroids = np.array([[0, 0]], dtype=np.int64)
    labels = np.array([0, 0], dtype=np.int64)

    dist = ng(X, centroids, labels)
    expected = np.array([[0.5], [0.5]])
    np.testing.assert_array_almost_equal(dist, expected)


def test_ng_requires_labels(encoded_data, centroids):
    # error if labels missing
    with pytest.raises(ValueError):
        distance(encoded_data, centroids, DistanceMetrics.NG, labels=None)


def test_distance_dispatcher(encoded_data, centroids, labels):
    # hamming dispatch
    d_h = distance(encoded_data, centroids, DistanceMetrics.HAMMING)
    assert d_h.shape == (encoded_data.shape[0], centroids.shape[0])

    # ng dispatch
    d_n = distance(encoded_data, centroids, DistanceMetrics.NG, labels=labels)
    assert d_n.shape == (encoded_data.shape[0], centroids.shape[0])


def test_ng_empty_cluster_handling():
    # handle empty cluster gracefully
    X = np.array([[0]], dtype=np.int64)
    centroids = np.array([[0], [1]], dtype=np.int64)
    labels = np.array([0], dtype=np.int64)  # c1 empty

    dist = ng(X, centroids, labels)
    # dist to c1 should be max (1.0 for 1 feature)
    assert dist[0, 1] == 1.0


def test_dimensions_mismatch(encoded_data, centroids):
    # error on dimension mismatch
    bad_centroids = centroids[:, :-1]
    with pytest.raises(ValueError, match="features"):
        hamming(encoded_data, bad_centroids)



import numpy as np
from kluster_fudge.utils import update_centroids


def test_update_centroids():
    # 2 clusters, 2 features
    # c0: [0,0], [0,1], [0,0] -> mode [0, 0]
    # c1: [1,1], [1,1], [2,1] -> mode [1, 1]

    X = np.array([[0, 0], [0, 1], [0, 0], [1, 1], [1, 1], [2, 1]], dtype=np.int64)

    labels = np.array([0, 0, 0, 1, 1, 1], dtype=np.int64)
    n_clusters = 2

    new_centroids = update_centroids(X, labels, n_clusters)

    expected = np.array([[0, 0], [1, 1]], dtype=np.int64)

    np.testing.assert_array_equal(new_centroids, expected)


def test_update_centroids_empty_cluster():
    # handle empty cluster (should return 0-vec)

    X = np.array([[10]], dtype=np.int64)
    labels = np.array([0], dtype=np.int64)  # c1 empty
    n_clusters = 2

    new_centroids = update_centroids(X, labels, n_clusters)

    # c1 should be 0 (default)
    assert new_centroids[1, 0] == 0



import numpy as np
import numpy.typing as npt
from numba import prange, njit


@njit(parallel=True, fastmath=True)
def update_centroids(
    X: npt.NDArray[np.int64], labels: npt.NDArray[np.int64], n_clusters: int
) -> npt.NDArray[np.int64]:
    """
    Compute new centroids by mode for each cluster for each feature.

    Args:
        X: (npt.NDArray[np.int64]) Encoded data array (n_samples, n_features)
        labels: (npt.NDArray[np.int64]) Cluster labels for each sample (n_samples, 1)
        n_clusters: (int) Number of clusters

    Returns:
        new_centroids: (npt.NDArray[np.int64]) Array of new centroids (n_clusters, n_features)
    """

    n_samples, n_features = X.shape
    new_centroids = np.zeros((n_clusters, n_features), dtype=np.int64)

    # parallelize over cols; note that this is thread safe bc each thread writes to diff col of new_centroids
    for j in prange(n_features):
        col = X[:, j]

        # find max val to allocate counts array size
        max_val = -1
        for i in range(n_samples):
            if col[i] > max_val:
                max_val = col[i]

        # allocate counts array
        counts = np.zeros((n_clusters, max_val + 1), dtype=np.int32)

        # count frequencies by point
        for i in range(n_samples):
            cluster_id = labels[i]
            val = col[i]
            counts[cluster_id, val] += 1

        # find mode for each row
        for k in range(n_clusters):
            best_val = 0
            best_count = -1
            for v in range(max_val + 1):
                if counts[k, v] > best_count:
                    best_count = counts[k, v]
                    best_val = v

            new_centroids[k, j] = best_val

    return new_centroids



import numpy as np
import pytest
import pandas as pd
from kluster_fudge.core import KModes


def test_init_validation():
    # invalid init method raises error
    with pytest.raises(ValueError, match="Unknown init method"):
        KModes(init_method="invalid")

    # invalid dist metric raises error
    with pytest.raises(ValueError, match="Unknown distance metric"):
        KModes(dist_metric="invalid")


def test_fit_predict_hamming(categorical_data):
    # init model
    model = KModes(n_clusters=3, dist_metric="hamming", max_iter=5)
    model.fit(categorical_data)

    # labels exist and match len
    assert model.labels is not None
    assert len(model.labels) == len(categorical_data)

    # check predict output shape
    preds = model.predict(categorical_data)
    assert preds.shape == (len(categorical_data),)


def test_fit_predict_jaccard(categorical_data):
    # jaccard metric runs ok
    model = KModes(n_clusters=3, dist_metric="jaccard", max_iter=5)
    model.fit(categorical_data)
    assert model.labels is not None


def test_fit_predict_ng(categorical_data):
    # ng metric runs ok
    model = KModes(n_clusters=3, dist_metric="ng", max_iter=5)
    model.fit(categorical_data)

    # labels exist
    assert model.labels is not None
    # should have >1 cluster
    assert len(np.unique(model.labels)) > 1


def test_perfect_separation():
    # create distinct groups
    # group 1: all 0s
    g1 = np.zeros((10, 4), dtype=int)
    # group 2: all 1s
    g2 = np.ones((10, 4), dtype=int)

    X = np.vstack([g1, g2])

    # fit model
    model = KModes(n_clusters=2, max_iter=10, init_method="random")
    model.fit(X)

    # first 10 same label
    assert len(np.unique(model.labels[:10])) == 1
    # last 10 same label
    assert len(np.unique(model.labels[10:])) == 1
    # distinct labels between groups
    assert model.labels[0] != model.labels[10]


def test_pandas_input(categorical_data):
    # dataframe input works
    df = pd.DataFrame(categorical_data, columns=["c1", "c2", "c3", "c4"])
    model = KModes(n_clusters=3)
    model.fit(df)

    # is_df flag set
    assert model.is_df
    assert model.labels is not None

    # predict works on df
    preds = model.predict(df)
    assert len(preds) == len(df)


def test_fit_predict_return(categorical_data):
    # fit_predict returns labels
    model = KModes(n_clusters=3)
    labels = model.fit_predict(categorical_data)

    assert labels is not None
    assert np.array_equal(labels, model.labels)


@pytest.mark.parametrize("method", ["random", "huang", "cao"])
def test_init_methods(categorical_data, method):
    # all init methods work
    model = KModes(n_clusters=3, init_method=method, max_iter=2)
    model.fit(categorical_data)
    assert model.centroids is not None


def test_determinism(categorical_data):
    # runs with same seed produce same results
    m1 = KModes(n_clusters=3, random_state=42)
    m1.fit(categorical_data)

    m2 = KModes(n_clusters=3, random_state=42)
    m2.fit(categorical_data)

    assert np.array_equal(m1.labels, m2.labels)
    assert np.array_equal(m1.centroids, m2.centroids)

    # different seed might produce same results on small data, so skipping the negative check


def test_simple_clusters():
    # easy clusters:
    # c0: all 0s
    # c1: all 1s
    # c2: all 2s

    c0 = np.zeros((10, 4), dtype=int)
    c1 = np.ones((10, 4), dtype=int)
    c2 = np.full((10, 4), 2, dtype=int)

    X = np.vstack([c0, c1, c2])

    # shuffle rows to make it slightly harder (but keeping structure)
    # actually let's keep it ordered to easily check ranges, model doesn't care about order

    model = KModes(n_clusters=3, init_method="cao")
    model.fit(X)

    # check that points 0-9 are same, 10-19 same, 20-29 same
    l0 = model.labels[:10]
    l1 = model.labels[10:20]
    l2 = model.labels[20:]

    assert len(np.unique(l0)) == 1
    assert len(np.unique(l1)) == 1
    assert len(np.unique(l2)) == 1

    # distinct groups
    assert l0[0] != l1[0]
    assert l1[0] != l2[0]
    assert l0[0] != l2[0]



import numpy as np
import numpy.typing as npt
from numba import prange, njit
from enum import Enum


class DistanceMetrics(Enum):
    HAMMING = "hamming"
    JACCARD = "jaccard"
    NG = "ng"


# we have list of centroid which we want to compare our input to
# a point is a list of xyz, in higher dimensions it has n items
# we have a list of centroids we want to compare against a list of targets


# X is data, Centroids is centroids
@njit(parallel=True, fastmath=True)
def hamming(
    X: npt.NDArray[np.int64], centroids: npt.NDArray[np.int64]
) -> npt.NDArray[np.float64]:
    """
    Compute Hamming distance between X and centroids.

    Args:
        X: (npt.NDArray[np.int64]) Data array (n_samples, n_features)
        centroids: (npt.NDArray[np.int64]) Centroids array (n_clusters, n_features)

    Returns:
        (npt.NDArray[np.float64]) Distance matrix (n_samples, n_clusters)
    """

    if X.shape[1] != centroids.shape[1]:
        raise ValueError("X and centroids must have the same number of features")

    rows = X.shape[0]  # NUMBER of rows
    # number of columns of distance matrix = number of centroids
    n_clusters = centroids.shape[0]

    # dist matrix
    distance: npt.NDArray[np.float64] = np.zeros((rows, n_clusters), dtype=np.float64)
    for i in prange(rows):  # iterate from 0 through the number of rows
        for j in range(n_clusters):
            dist = 0
            for a, b in zip(X[i], centroids[j]):
                if a != b:
                    dist += 1
            distance[i][j] = dist
    return distance


@njit(parallel=True, fastmath=True)
def jaccard(
    X: npt.NDArray[np.int64], centroids: npt.NDArray[np.int64]
) -> npt.NDArray[np.float64]:
    """
    Compute Jaccard distance between X and centroids.

    Args:
        X: (npt.NDArray[np.int64]) Data array (n_samples, n_features)
        centroids: (npt.NDArray[np.int64]) Centroids array (n_clusters, n_features)

    Returns:
        (npt.NDArray[np.float64]) Distance matrix (n_samples, n_clusters)
    """

    if X.shape[1] != centroids.shape[1]:
        raise ValueError("X and centroids must have the same number of features")

    rows, cols = X.shape
    # number of columns of distance matrix = number of centroids
    n_clusters = centroids.shape[0]

    # dist matrix
    distance: npt.NDArray[np.float64] = np.zeros((rows, n_clusters), dtype=np.float64)
    for i in prange(rows):  # iterate from 0 through the number of rows
        for j in range(n_clusters):
            dist = 0
            for a, b in zip(X[i], centroids[j]):
                if a == b:
                    dist += 1
            distance[i][j] = (
                1 - (dist / (cols * 2 - dist))
            )  # double the num of columns (because a union between both centroids and data points), then subtract the numner of similar elements
    return distance


@njit(parallel=True, fastmath=True)
def ng(
    X: npt.NDArray[np.int64],
    centroids: npt.NDArray[np.int64],
    labels: npt.NDArray[np.int64],
) -> npt.NDArray[np.float64]:
    """
    Compute NG Distance between X and centroids.

    Args:
        X: (npt.NDArray[np.int64]) Data array (n_samples, n_features)
        centroids: (npt.NDArray[np.int64]) Centroids array (n_clusters, n_features)
        labels: (npt.NDArray[np.int64]) Labels array (n_samples,)

    Returns:
        (npt.NDArray[np.float64]) Distance matrix (n_samples, n_clusters)

    """
    if X.shape[1] != centroids.shape[1]:
        raise ValueError("X and centroids must have the same number of features")

    rows, cols = X.shape
    n_clusters = centroids.shape[0]

    # find max value to determine array size for counts
    max_val = 0
    for r in range(rows):
        for c in range(cols):
            if X[r, c] > max_val:
                max_val = X[r, c]

    # allocate counts & cluster sizes
    counts = np.zeros((n_clusters, cols, max_val + 1), dtype=np.int32)
    cluster_sizes = np.zeros(n_clusters, dtype=np.int32)

    # populate counts & cluster sizes
    for r in range(rows):
        cluster_id = labels[r]
        if cluster_id < 0 or cluster_id >= n_clusters:
            continue  # skip invalid labels (shouldn't happen)
        cluster_sizes[cluster_id] += 1
        for c in range(cols):
            val = X[r, c]
            counts[cluster_id, c, val] += 1

    distance = np.zeros((rows, n_clusters), dtype=np.float64)

    for r in prange(rows):
        for i in range(n_clusters):
            dist = 0.0
            if cluster_sizes[i] == 0:
                # If cluster is empty, max distance
                dist = float(cols)
            else:
                for c in range(cols):
                    val = X[r, c]
                    if val <= max_val:
                        freq = counts[i, c, val]
                        prob = freq / cluster_sizes[i]
                        # given by: sum(1 - P(x_j | C))
                        dist += 1.0 - prob
                    else:
                        dist += 1.0  # Should not happen based on max_val logic

            distance[r, i] = dist

    return distance


def distance(
    X: np.ndarray,
    centroids: np.ndarray,
    metric: DistanceMetrics,
    labels: npt.NDArray[np.int64] | None = None,
) -> npt.NDArray[np.float64]:
    """
    Compute distance between X and centroids using the specified metric.

    Args:
        X: (npt.NDArray[np.int64]) Data array (n_samples, n_features)
        centroids: (npt.NDArray[np.int64]) Centroids array (n_clusters, n_features)
        metric: (DistanceMetrics) Distance metric to use
        labels: (npt.NDArray[np.int64] | None) Labels array (n_samples,) for ng dist

    Returns:
        (npt.NDArray[np.float64]) Distance matrix (n_samples, n_clusters)
    """

    if metric == DistanceMetrics.HAMMING:
        return hamming(X, centroids)
    elif metric == DistanceMetrics.JACCARD:
        return jaccard(X, centroids)
    elif metric == DistanceMetrics.NG:
        if labels is None:
            raise ValueError("Labels are required for NG distance")
        return ng(X, centroids, labels)
    else:
        raise ValueError(f"Unsupported distance metric: {metric}")



from enum import Enum
import numpy.typing as npt
import numpy as np
from numba import njit, prange
from kluster_fudge.dist import distance, DistanceMetrics


class InitMethod(Enum):
    RAND = 0
    HUANG = 1
    CAO = 2


def init_centroids(
    X: npt.NDArray[np.float64],
    n_clusters: int,
    method: InitMethod,
    random_state: int | None = None,
) -> np.ndarray:
    """
    Initialize centroids using the specified method.

    Args:
        X: (npt.NDArray[np.float64]) Data array (n_samples, n_features)
        n_clusters: (int) Number of clusters
        method: (InitMethod) Initialization method
        random_state: (int | None) Random seed for reproducibility

    Returns:
        (npt.NDArray[np.float64]) Centroids array (n_clusters, n_features)
    """
    if random_state is not None:
        np.random.seed(random_state)

    if method == InitMethod.RAND:
        n_samples = X.shape[0]
        if n_clusters > n_samples:
            raise ValueError(
                f"n_clusters ({n_clusters}) cannot exceed n_samples ({n_samples})"
            )

        random_indices = np.random.choice(n_samples, n_clusters, replace=False)
        return X[random_indices]
    elif method == InitMethod.HUANG:
        return _init_centroids_huang(X, n_clusters)
    elif method == InitMethod.CAO:
        return _init_centroids_cao(X, n_clusters)
    else:
        raise ValueError(f"Unknown init method: {method}")


def _freq_sort_categories(X: npt.ArrayLike) -> np.ndarray:
    """
    Sort the unique values in X by frequency in descending order.

    Args:
        X: (npt.ArrayLike) Data array (n_samples,)

    Returns:
        (npt.NDArray[np.float64]) Sorted unique values
    """

    unique_vals, counts = np.unique(X, return_counts=True)
    sort_indices = np.argsort(counts)
    return unique_vals[sort_indices[::-1]]


def _init_centroids_huang(X: npt.ArrayLike, n_clusters: int) -> np.ndarray:
    """
    Initialize centroids using the Huang algorithm.

    Args:
        X: (npt.ArrayLike) Data array (n_samples, n_features)
        n_clusters: (int) Number of clusters

    Returns:
        (npt.NDArray[np.float64]) Centroids array (n_clusters, n_features)
    """

    X = np.asarray(X)
    _, n_features = X.shape

    if X.ndim != 2:
        raise ValueError("Input array must be 2-dimensional")

    sorted_cols = [_freq_sort_categories(X[:, j]) for j in range(n_features)]

    synthetic_centroids = np.empty((n_clusters, n_features), dtype=X.dtype)

    for i in range(n_clusters):
        for j in range(n_features):
            unique_vals = sorted_cols[j]
            synthetic_centroids[i, j] = unique_vals[i % len(unique_vals)]

    final_centroids = np.empty((n_clusters, n_features), dtype=X.dtype)
    taken_indices = set()

    for i in range(n_clusters):
        best_idx = -1
        distances = np.sum(X != synthetic_centroids[i], axis=1)
        candidate_indices = np.argsort(distances)

        for idx in candidate_indices:
            if idx not in taken_indices:
                best_idx = idx
                break

        if best_idx == -1:
            best_idx = candidate_indices[0]

        taken_indices.add(best_idx)
        final_centroids[i] = X[best_idx]

    return final_centroids


@njit(parallel=True, fastmath=True)
def _compute_X_density(X: npt.ArrayLike) -> np.ndarray:
    """
    Compute the density of each data point in X.

    Args:
        X: (npt.ArrayLike) Data array (n_samples, n_features)

    Returns:
        (npt.NDArray[np.float64]) Density array (n_samples,)
    """

    X = np.asarray(X)
    U, A = X.shape

    densities = np.zeros(U, dtype=np.float64)

    for a in range(A):
        col = X[:, a]
        max_val = np.max(col)
        counts = np.zeros(max_val + 1, dtype=np.int64)
        for u in range(U):
            counts[col[u]] += 1

        for u in prange(U):
            val = col[u]
            densities[u] += counts[val]

    return densities / (U * A)


def _init_centroids_cao(X: npt.NDArray[np.float64], n_clusters: int) -> np.ndarray:
    """
    Initialize centroids using the Cao algorithm.

    Args:
        X: (npt.NDArray[np.float64]) Data array (n_samples, n_features)
        n_clusters: (int) Number of clusters

    Returns:
        (npt.NDArray[np.float64]) Centroids array (n_clusters, n_features)
    """

    densities = _compute_X_density(X)
    first_centroid_index = np.argmax(densities)
    centroid_set = [X[first_centroid_index]]
    n_samples, _ = X.shape
    min_dists = distance(
        X, X[first_centroid_index].reshape(1, -1), DistanceMetrics.HAMMING
    ).flatten()

    for _ in range(n_clusters - 1):
        # compute score
        scores = densities * min_dists
        best_idx = np.argmax(scores)
        new_centroid = X[best_idx]
        centroid_set.append(new_centroid)

        # compute distance to new centroid
        dist_to_new = distance(
            X, new_centroid.reshape(1, -1), DistanceMetrics.HAMMING
        ).flatten()
        min_dists = np.minimum(min_dists, dist_to_new)

    return np.array(centroid_set)
